/**
 * Copyright (c) 2020 Raspberry Pi (Trading) Ltd.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 */

#include <stdio.h>
#include <math.h>
#include <string.h>

#if PICO_ON_DEVICE

#include "hardware/clocks.h"
#include "hardware/structs/clocks.h"

#endif

#include "pico/stdlib.h"

#if USE_AUDIO_I2S

#include "pico/audio_i2s.h"

#if PICO_ON_DEVICE
#include "pico/binary_info.h"
bi_decl(bi_3pins_with_names(PICO_AUDIO_I2S_DATA_PIN, "I2S DIN", PICO_AUDIO_I2S_CLOCK_PIN_BASE, "I2S BCK", PICO_AUDIO_I2S_CLOCK_PIN_BASE+1, "I2S LRCK"));
#endif

#endif

#include "fatfs/ff.h"

#define SINE_WAVE_TABLE_LEN 2048
#define SAMPLES_PER_BUFFER 256

#if PICO_ON_DEVICE == 1
#warning PICO_ON_DEVICE is 1
#elif PICO_ON_DEVICE == 0
#warning PICO_ON_DEVICE is 0
#else
#warning PICO_ON_DEVICE is X
#endif

static int16_t sine_wave_table[SINE_WAVE_TABLE_LEN];

struct audio_buffer_pool *init_audio(uint32_t sample_freq, uint16_t channel_count, uint16_t sample_stride) {

    static audio_format_t audio_format = {
            .format = AUDIO_BUFFER_FORMAT_PCM_S16,
#if WAV_PLAYER
//            .sample_freq = 44100,
            .sample_freq = 24000,
#else
            .sample_freq = 24000,
#endif
#if PICO_AUDIO_I2S_MONO_INPUT            
            .channel_count = 1,		// 6/17
#else
            .channel_count = 2,
#endif
    };

    static struct audio_buffer_format producer_format = {
            .format = &audio_format,
#if PICO_AUDIO_I2S_MONO_INPUT
            .sample_stride = 2		// 6/17
#else
            .sample_stride = 4
#endif
    };

	audio_format.sample_freq = sample_freq;
	audio_format.channel_count = channel_count;
	producer_format.sample_stride = sample_stride;
	
    struct audio_buffer_pool *producer_pool = audio_new_producer_pool(&producer_format, 3,
                                                                      SAMPLES_PER_BUFFER); // todo correct size
    bool __unused ok;
    const struct audio_format *output_format;
#if USE_AUDIO_I2S
    struct audio_i2s_config config = {
            .data_pin = 18,	//PICO_AUDIO_I2S_DATA_PIN,
            .clock_pin_base = 16,	//PICO_AUDIO_I2S_CLOCK_PIN_BASE,
            .dma_channel = 0,
            .pio_sm = 0,
    };

    output_format = audio_i2s_setup(&audio_format, &config);
    if (!output_format) {
        panic("PicoAudio: Unable to open audio device.\n");
    }

    ok = audio_i2s_connect(producer_pool);		// この中でconsumer_poolを生成
    assert(ok);
    audio_i2s_set_enabled(true);
#endif
    return producer_pool;
}

int main() {
    FATFS fs;
    FIL fil;
    FRESULT fr;     /* FatFs return code */
    UINT br;
    UINT bw;
    uint8_t		wavheader[64];
    uint32_t	sample_freq;
    uint16_t	channel_count;
	uint16_t	bits_per_sample;
	uint16_t	sample_stride;
	
    stdio_init_all();

	// FatFs
    fr = f_mount(&fs, "", 1);
    if (fr != FR_OK) {
        printf("mount error %d\n", fr);
        return 0;
    }
    printf("mount ok\n");
    switch (fs.fs_type) {
        case FS_FAT12:
            printf("Type is FAT12\n");
            break;
        case FS_FAT16:
            printf("Type is FAT16\n");
            break;
        case FS_FAT32:
            printf("Type is FAT32\n");
            break;
        case FS_EXFAT:
            printf("Type is EXFAT\n");
            break;
        default:
            printf("Type is unknown\n");
            break;
    }
    printf("Card size: %7.2f GB (GB = 1E9 bytes)\n\n", fs.csize * fs.n_fatent * 512E-9);
    fr = f_open(&fil, "wavplay.wav", FA_READ);
    if (fr != FR_OK) {
        printf("open error %d\n", fr);
        return 0;
    }
    f_read(&fil, wavheader, 0x2c, &br);
    memcpy(&channel_count, &wavheader[0x16], 2);
    memcpy(&sample_freq, &wavheader[0x18], 4);
    memcpy(&bits_per_sample, &wavheader[0x22], 2);
	sample_stride = channel_count * bits_per_sample / 8;
	
    for (int i = 0; i < SINE_WAVE_TABLE_LEN; i++) {
        sine_wave_table[i] = 32767 * cosf(i * 2 * (float) (M_PI / SINE_WAVE_TABLE_LEN));
    }

    struct audio_buffer_pool *ap_producer = init_audio(sample_freq, channel_count, sample_stride);		// producer
    puts("init_audio\n");		// 6/17
    uint32_t step = 0x200000;
    uint32_t pos = 0;
    uint32_t pos_max = 0x10000 * SINE_WAVE_TABLE_LEN;
    uint vol = 10;		// 128;
    while (true) {
        int c = getchar_timeout_us(0);
        if (c >= 0) {
            if (c == '-' && vol) vol -= 4;
            if ((c == '=' || c == '+') && vol < 255) vol += 4;
            if (c == '[' && step > 0x10000) step -= 0x10000;
            if (c == ']' && step < (SINE_WAVE_TABLE_LEN / 16) * 0x20000) step += 0x10000;
            if (c == 'q') break;
            printf("vol = %d, step = %d      \r", vol, step >> 16);
        }
        struct audio_buffer *buffer = take_audio_buffer(ap_producer, true);		// pico-extras\...\pico_audio\audio.cpp
        printf("audio_buffer/p (%d)\n", buffer->bufferno);
        int16_t *samples = (int16_t *) buffer->buffer->bytes;
#if WAV_PLAYER
		fr = f_read(&fil, samples, buffer->max_sample_count * 4, &br);	// 1sample=4byte 焼き付け
		if (fr != FR_OK || br == 0) break;
        for (uint i = 0; i < buffer->max_sample_count; i++) {
            samples[2*i] = (vol * samples[2*i]) >> 8u;
            samples[2*i + 1] = (vol * samples[2*i + 1]) >> 8u;
        }
#else
        for (uint i = 0; i < buffer->max_sample_count; i++) {
#if PICO_AUDIO_I2S_MONO_INPUT
            samples[i] = (vol * sine_wave_table[pos >> 16u]) >> 8u;		// 6/17
#else
            samples[2*i] = (vol * sine_wave_table[pos >> 16u]) >> 8u;
            samples[2*i + 1] = samples[2*i];
#endif
            pos += step;
            if (pos >= pos_max) pos -= pos_max;
        }
#endif
        //printf("take_audio_buffer p=%x  %x  %x  %x\n", (uint32_t)buffer, (uint32_t)buffer->buffer, (uint32_t)buffer->buffer->bytes, *(uint32_t *)buffer->buffer->bytes);
        buffer->sample_count = buffer->max_sample_count;
        give_audio_buffer(ap_producer, buffer);
    }
    puts("end \n");
    return 0;
}
